#include "defo.h"
#include "estimates-functions.h"
#include "bodies.h"
#include "cuteCalc.h"



bool findCluster(const Bodies &bodies, int selBody, MyDouble delta, vector<int> &cluster)
/**
We look for the cluster containg the body <selBody>
Input: 
  bodies - box in configuration space containing bodies  * Numbering of bodies is zero-based (0, 1, ..., bodies.N - 1).
  num - the first body in the cluster
  delta - the body is in the cluster, if there exists a chain of bodies connecting it to <selBody>
     and each distance in the chain is less than <delta>
On output: 
      In the `cluster` vector we have indices of the bodies from the cluster.
 * Returns `false` if the cluster is a singleton (i.e. the <selBody>-th body is isolated,
 * meaning that there're no other bodies within the delta radius).
 */
{
    int number_of_bodies=bodies.N;
    bool result = false;
    int j;
    vector<int> tempT;
    // marker array: markedBodies[i] = 1 <=> i-th body is already in the cluster
    int markedBodies[number_of_bodies];

    // set all zeros in the `cluster` vector
    cluster.clear();

    for (int i = 0; i < number_of_bodies; i++)
        markedBodies[i] = 0;

    tempT.push_back(selBody);
    cluster.push_back(selBody);
    markedBodies[selBody] = 1;

    do
    {
        j = tempT.back();
        tempT.pop_back();  // we remove the last entry in <tempT>
        for (int i = 0; i < number_of_bodies; i++)
        {
            if (i == j)
                continue;

            if (markedBodies[i] == 1) // don't add a body if it's already added
                continue;

            if (bodies.radius(i, j).leftBound() < delta)
            {
                cluster.push_back(i);
                tempT.push_back(i);
                markedBodies[i] = 1;
                result = true;  // at least one body (different from <selBody>) has been added to the cluster
            }
        }
    } while (!tempT.empty());

    return result;
}

MyInterval supMomInertiaCluster(const Bodies &bodies, vector<int> &cluster)
// Returns  I_{C,Z} \geq \sum_{i\in cluster} m_i * q_i^2
{
    MyInterval sum = 0.0;
    const int vectorSize = cluster.size();
    int i;

    for (int j = 0; j < vectorSize; j++)
    {
        i = cluster[j];
		#ifdef PLANAR
        sum += bodies.mass(i) * (power(bodies.x(i), 2) + power(bodies.y(i), 2));
		#endif
		#ifdef SPATIAL
		sum += bodies.mass(i) * (power(bodies.x(i), 2) + power(bodies.y(i), 2) + power(bodies.z(i), 2));
		#endif
    }

    return sum.rightBound();
}

MyInterval infPotEnergyCluster(const Bodies &bodies, vector<int> &cluster)
// returns U_{C,Z} = inf \sum_{i < j\in cluster} \frac{m_i*m_j}{r_{i, j}} 
{
    MyInterval rTemp;
    const int vectorSize = cluster.size();
    int iIndex, jIndex;

    MyInterval infC = 0.0;
    for (int i = 0; i < vectorSize; i++)
    {
        iIndex = cluster[i];
        for (int j = i + 1; j < vectorSize; j++)
        {
            jIndex = cluster[j];
            rTemp = bodies.radius(iIndex, jIndex);

            infC += (bodies.mass(iIndex).leftBound() * bodies.mass(jIndex).leftBound()) / rTemp.rightBound();
        }
    }

    return infC;
}



void findClusterComplement(int numBodies, vector<int> &cluster, vector<int> &clusterCompl)
/* In the `clusterCompl` we store the indices of bodies not in the cluster.
  Numbering of the bodies is zero-based (0, 1, ..., numBodies-1)  */
{
    int j;
    const int clusterSize = cluster.size();
    bool iFound = false;

    clusterCompl.clear();

    for (int i = 0; i < numBodies; i++)
    {
        iFound = false;
        j = 0;
        do
        {
            if (i == cluster[j]) // i-th body is in the cluster
                iFound = true;
            j++;
        } while ((j < clusterSize) && !iFound);

        if (!iFound) // i-th body is not in the cluster
            clusterCompl.push_back(i);
    }
}

MyInterval infFCZ(const Bodies &bodies, vector<int> &cluster)
/*  F_{C,Z}(q) = \sum_{i\in cluster, k\not\in cluster} \frac{m_i * m_k}{r_{i, k}^3} (q_i - q_k|q_i)
 * */
{
    MyInterval  sPTemp, bMass, infTemp;
    vector<int> clusterCompl;
    MyInterval infF = 0.0;

    const int numBodies=bodies.N;

    const int vectorSize = cluster.size();
    if (vectorSize != numBodies)
    // If the cluster contains all bodies, then we set infF = 0.0 and stop.
    {
        int iIndex, jIndex;

        infTemp = 0.0;
        findClusterComplement(numBodies,cluster, clusterCompl);
        int complementSize = clusterCompl.size();
        for (int i = 0; i < vectorSize; i++)
        {
            iIndex = cluster[i];
            for (int j = 0; j < complementSize; j++)
            {
                jIndex = clusterCompl[j];
				#ifdef PLANAR
                MyInterval qx, qy;
			    MyInterval distx = bodies.x(iIndex)-bodies.x(jIndex);
                MyInterval disty = bodies.y(iIndex)-bodies.y(jIndex);
                cuteCalculation(distx, disty, qx, qy, 3, 1);
				 // sPTemp = (q_{iIndex} - q_{jIndex}|q_{iIndex})
                sPTemp = bodies.x(iIndex) * qx + bodies.y(iIndex) * qy;
                #endif
				#ifdef SPATIAL
				MyInterval qx, qy, qz;
                MyInterval distx = bodies.x(iIndex) - bodies.x(jIndex);
                MyInterval disty = bodies.y(iIndex) - bodies.y(jIndex);
                MyInterval distz = bodies.z(iIndex) - bodies.z(jIndex);
                cuteCalculation(distx, disty, distz, qx, qy, qz, 3, 1);

                // sPTemp = (q_{iIndex} - q_{jIndex}|q_{iIndex})
                sPTemp = bodies.x(iIndex) * qx + bodies.y(iIndex) * qy + bodies.z(iIndex) * qz;
				#endif
                bMass = bodies.mass(iIndex) * bodies.mass(jIndex);

                infTemp += bMass * sPTemp;
            }
        }
        infF = infTemp.leftBound();
    }

    return infF;
}

bool isInCluster(vector<int> &cluster, int i)
/** Returns `true` if the `i`-th body is in the cluster. */
{
    const int s = cluster.size();
    for (int j = 0; j < s; j++)
        if (cluster[j] == i)
            return true;

    return false;
}
