#include "defo.h"

#include "bodies.h"
#include "krawczyk-newton.h"
#include "MyIntervalFunc.h"
#include "test-and-help-functions.h"
#include "searchalg.h"

#include "cc-nb.h"


/**
 * The asynchronous search:
 * The search fires recursively on both parts of the subdivided interval
 * unless the stopping conditions are met. Each recursive call tracks the
 * adds one to the global count of parallel searchers running. We don't
 * perform the call asynchronously if certain threshold has been reached.
 * This is because we want to avoid the overhead of starting to many threads
 * on the host system. Note that once the asynchronous search function finishes,
 * it adds back to the pool of available threads, so later subdvisions can again
 * be performed asynchronously.
 */
Count searchCC(Bodies bodies, bool searchStage,ostream &fhsolfound, ostream &fhboxfailed) 
//  it uses some global ostreams to output binary data:
//  solutions (boxes) found are written to   <fhsolfound>, the undecided boxes to <fhboxfailed>
{
    if (searchStage)
        boxCount++;   // the global variable counting all boxes treated
 
    ResultType newtonRes = ResultType::uniqueZero;
    Count result = {0, 0};
    Count leftCount, rightCount;
    std::future<Count> leftCountF, rightCountF;
    bool shouldAsync = cnt < MAX_ASYNC_DEPTH;
	#ifdef NO_PARALLEL_COMP
	shouldAsync=false;
	#endif
  

    if (searchStage)
    {
        if (!possibleSolution(bodies))  // bodies - can be updated (shrinked) in the execution of this function 
		                                                  // so that all possible CC in bodies (on input) are contained in bodies(on output)
        {
		   result.zeros = 0;
           result.maybeZeros = 0;
           return result;
        }
    }

  
    bool collision = bodies.collisionHappened();

    // We're looking for the longest interval to subdivide.
    // The found interval describes the max_it-th body.
    // coord denotes whether to subdivide on x, y or z coord.
    int max_i = 0;
    int coord = 0; // coord = 0 => subdivide x, coord = 1 => subdivide y, coord = 2 => subdivide z
    MyDouble maxLengthInterval = bodies.maxDiam(max_i, coord);

    bool intersec = false;  // when true then we know that we got an improvement in Newton/Krawczyk method

    if (!collision)
	{
        if (maxLengthInterval < bias)  // bias is a global variable, we call the costly Nowton/Krawczyk method, when  the size of the set is <= eps (another global variable)
        {
            // possible zero;
            // use Newton's/Krawczyk's method to check if that's true:
            Bodies bodiesRetFromKraw(bodies.N);
			int dimRCS=bodies.dimRedConfSpace();
            MyMatrix DFF(dimRCS, dimRCS);
            computeDFRedMatrix(bodies, DFF);
			#ifdef FINITENESS_TEST    
            if (isomorphismTest(DFF))
            {
                result.zeros = 1;
                // if the program is run in the mode of testing finiteness of the number of CC's, then zeros = 1 means that
                // there is at most 1 zeros (zeros is upper bound of the number of CC's)
                result.maybeZeros = 0;
                isoBoxesCount++;
                //cout << "isoTest nr" << isoBoxesCount << " jest kostka nr " << boxCount << endl
                //     << flush;
                return result;
            }
			#endif

            newtonRes = (methodUsed == ZeroSearchType::newton) ? newtonMethod(bodies, bodiesRetFromKraw, intersec, DFF) : krawczykMethod(bodies, bodiesRetFromKraw, intersec, DFF);

          

            switch (newtonRes)
            {
            case ResultType::methodFailed:
                if (searchStage)
                    KrawczykFailedCounter++;
                if (intersec)
				{
                    bodies = bodiesRetFromKraw;   // we gained something, the size is smaller
					bodies.maxDiam(max_i, coord);  // we update max_i and coord because <bodies> has been changed
				}	
                break; 

            case ResultType::uniqueZero:

                if (searchStage)
                    KrawczykZeroCounter++;
                methodZeroes++;
				#ifndef FINITENESS_TEST // FINITENESS_TEST, then we just count the number of possible zero's hence we skip this section
                krawczykSection_mutex.lock();
                bodiesRetFromKraw.writeBinaryData(fhsolfound); // 
                krawczykSection_mutex.unlock();
				#endif

                result.zeros = 1;
                result.maybeZeros = 0;
                return result;

            case ResultType::noZeroInSet:
                result.zeros = 0;
                result.maybeZeros = 0;
                if (searchStage)
                    KrawczykNoZeroCounter++;
                return result;

            default:
                break;
            }
        }	
	}
	
	
    if (maxLengthInterval < eps)
    // if the size of the cube is too small, then we interrupt search
    {
        if (searchStage)
        {                          
            methodFailedCounter++; // global variable counts undecided cubes!
			#ifndef FINITENESS_TEST            
            write_mutex.lock();
            bodies.writeBinaryData(fhboxfailed); // !
            write_mutex.unlock();
			#endif
        }
        result.zeros = 0;
        result.maybeZeros = 1;
        return result;
    }

    // we do binary division in direction of max_i coordinate
	MyDouble leftPoint = 0, rightPoint = 0, midPoint = 0, leftMidPoint = 0, rightMidPoint = 0;
    MyDouble overlap = 1e-6; // overlap has to be less than 0.5
	if(coord==0)  // x-variable
    {
	    leftPoint = bodies.x(max_i).leftBound();
        rightPoint = bodies.x(max_i).rightBound();  
	}
	if(coord==1) // y-variable
	{  
		leftPoint = bodies.y(max_i).leftBound();
        rightPoint = bodies.y(max_i).rightBound();
	}
	#ifdef SPATIAL
	if(coord==2) // z-variable
	{
		leftPoint = bodies.z(max_i).leftBound();
        rightPoint = bodies.z(max_i).rightBound();
	}
	#endif
	 midPoint = (leftPoint + rightPoint) / 2.0;
     rightMidPoint = midPoint + (rightPoint - leftPoint) * overlap;
     leftMidPoint = midPoint - (rightPoint - leftPoint) * overlap;
	
	
	
	// launching the computation on the "left-half" of bodies
    if(coord==0)  // x-variable
        bodies.setX(max_i, MyInterval(leftPoint, rightMidPoint));
	if(coord==1) // y-variable
        bodies.setY(max_i, MyInterval(leftPoint, rightMidPoint));
    #ifdef SPATIAL
    if(coord==2) // z-variable 
        bodies.setZ(max_i, MyInterval(leftPoint, rightMidPoint));
	#endif
   

    auto policy = std::launch::async; //| std::launch::deferred;
    if (shouldAsync)
    {
	    #ifndef NO_PARALLEL_COMP
        leftCountF = std::async(policy, [bodies, searchStage,&fhsolfound,&fhboxfailed]() { return searchCC(bodies, searchStage,fhsolfound,fhboxfailed); });
		#endif
		
        #ifdef NO_PARALLEL_COMP
		 cout << " cos sie nie zgadza z wylaczeniem rownoleglosci " << endl;
		 exit(1);
		#endif
        cnt += 1;
    }
    else
    {
        leftCount = searchCC(bodies, searchStage,fhsolfound,fhboxfailed);
    }

    // now on the "right-half" 
     if(coord==0)  // x-variable
        bodies.setX(max_i, MyInterval(leftMidPoint, rightPoint));
    if(coord==1) // y-variable
        bodies.setY(max_i, MyInterval(leftMidPoint, rightPoint));
	#ifdef SPATIAL
    if(coord==2) // z-variable 
        bodies.setZ(max_i, MyInterval(leftMidPoint, rightPoint));
	#endif
   
    if (shouldAsync)
    {  
	    #ifndef NO_PARALLEL_COMP
        rightCountF = std::async(policy, [bodies, searchStage,&fhsolfound,&fhboxfailed]() { return searchCC(bodies, searchStage,fhsolfound,fhboxfailed); });
		#endif
		
        #ifdef NO_PARALLEL_COMP
		 cout << " cos sie nie zgadza z wylaczeniem rownoleglosci " << endl;
		 exit(1);
		#endif
        leftCount = leftCountF.get();
        rightCount = rightCountF.get();
        cnt -= 1;  
    }
    else
    {
        rightCount = searchCC(bodies, searchStage,fhsolfound,fhboxfailed);
    }

    result.zeros = leftCount.zeros + rightCount.zeros;
    result.maybeZeros = leftCount.maybeZeros + rightCount.maybeZeros;
    return result;

} //search





bool blowUpAndProveUniqueSolution(Bodies bodies)
/* to be used on  <bodies> with low diameter,  we already know that there a CC in <bodies>;
  we want to prove that <bodies> contain a unique solution; 
  
  
  sometimes in interval arithmetic can happen that on the result from 
  the Krawczyk operator we cannot establish the existence , then we should apply this function
  for example when the solution was obtained in more optimised calculations  
  */
  
 //  returns true, when the Krawczyk/Newton method have shown that there is indeed unique solution in <bodies>
//     and false otherwise 
//    it write (appends) the established box on which we have a unique solution to <zeroTestedFile> 
{
  /* 
   
    in the loop we "puff" the boides and try to (dis)prove the existence of the solution in the set
     if not, the zero can still be present
   */
  
    bool isUniqueSol=false;
	int  dimRCS=bodies.dimRedConfSpace();

    MyMatrix  DFF(dimRCS,dimRCS);
	Bodies KrawRetval(bodies);  // this  will copy  masses  and other settings which my appear
    ResultType newtonRes = ResultType::methodFailed;
	
	
	
	
	 //  now we start puffing the bodies
    Bodies *midBodies = bodies.calculateMidBodies();
     int trash1 = 0, trash2 = 0;
     MyDouble maxLengthInterval = bodies.maxDiam(trash1, trash2);
     MyInterval half = MyInterval(maxLengthInterval) / 2.0;
     half = MyInterval(-half.rightBound(), half.rightBound());
     Bodies puffedBodies(bodies);  // it is important that puffedBodies are created as copy of <bodies>, bacause in this way the masses are copied and
	                               // maybe some other setting introduced in the future
     calculatePuffedBodies(*midBodies, half, puffedBodies);  // at the moment bodies \subset puffedBodies
     delete midBodies; 
    
    
     bool theEnd = false;
     while ((newtonRes != ResultType::uniqueZero) && (!theEnd))
     {

        // this shouldn't be the case, but check for collisions:
        bool collision = puffedBodies.collisionHappened();
        bool intersec; // this is ignored in the algorithm

         
        if (!collision)
        {
	        computeDFRedMatrix(puffedBodies, DFF);
            newtonRes = (methodUsed == ZeroSearchType::newton) ? newtonMethod(puffedBodies, KrawRetval, intersec,DFF) : krawczykMethod(puffedBodies, KrawRetval, intersec,DFF);

            if(newtonRes==ResultType::methodFailed) 
			{
			    maxLengthInterval = puffedBodies.maxDiam(trash1, trash2);
                if (maxLengthInterval >= 1e-2)
                    theEnd = true;
                else
                {
                  half = MyInterval(maxLengthInterval) / 2.0;
                  half = MyInterval(-half.rightBound(), half.rightBound());
                  calculatePuffedBodies(puffedBodies, half, puffedBodies);
                }
			}  
           
           
            if(newtonRes==ResultType::uniqueZero)
			{
               methodZeroes++;
           /*    ofstream outFile(zerosTestedFile, ios::binary | ios::out | ios::app);
               KrawRetval.writeBinaryData(outFile); 
               outFile.close(); */
               isUniqueSol=true;
			   theEnd=true;
			}
                    

            if(newtonRes==ResultType::noZeroInSet)
			{
               theEnd = true;
               cout << "No zeros in blowUpAndProveUniqueSolution - this should not happen" << endl;
			   cout << "Problematic box: " << endl;
			   bodies.printData();
			   bodies.printDiam();
            }
        }else theEnd=true;  // there is collision
    }
   
    return isUniqueSol;
}

